from saboteurs import find_logical_saboteurs

# NOTE: the teo definitions below could also be read from a CSV:
# groups, failed_groups = csv_to_groups_data('find_logical_saboteurs.csv')

groups = {
    1: ["A", "C", "D"],
    2: ["B", "C", "E"],
    3: ["A", "B", "D"],
    4: ["D", "F", "G"],
}
failed_groups = [2, 4]

result = find_logical_saboteurs(groups, failed_groups=failed_groups)
print(result)



from collections import OrderedDict
import itertools
import numpy as np
from .minimal_cover import minimal_cover


def generate_combinatorial_groups(elements_per_position, prefix="group_"):
    """Generate all possibilities from a combinatorial design.

    Parameters
    ----------
    elements_per_position : dict
        A dictionary where the key is 'position_name' and the value
        is a list of elements at this position.
    prefix : str
        Prefix to generate names for the generated groups,
        which will be of the form prefix_001, prefix_002, etc.

    Returns
    -------
    groups : collections.OrderedDict
        An ordered dictionary with keys formatted as 'group_001', 'group_002', etc.,
        each key corresponding to a list of elements, [e1, e2, e3, e4, ...].
        Each group contains exactly one element from each position,
        and the groups represent all possible combinations of elements.
    """
    if hasattr(elements_per_position, "values"):
        elements_per_position = list(elements_per_position.values())
    groups = list(itertools.product(*elements_per_position))
    n_zeros = int(np.log10(len(groups))) + 1
    return OrderedDict(
        [(prefix + str(i + 1).zfill(n_zeros), group) for i, group in enumerate(groups)]
    )


def _minimal_elements_group_coverage(groups):
    elements_group_coverage = {}
    for group_name, elements in groups.items():
        for element in elements:
            if element not in elements_group_coverage:
                elements_group_coverage[element] = set()
            elements_group_coverage[element].add(group_name)
    return minimal_cover(set(groups.keys()), elements_group_coverage.items())


def design_test_batch(possible_groups, max_saboteurs=1):
    """Select a subset of the groups that enables identification of bad elements.

    Parameters
    ----------
    possible_groups : dict
        A dictionary where the key is 'group_name' and the value is a list of elements in that group.
    max_saboteurs : int
        The maximum number of potential bad elements among all elements in all
        the groups. A bad element is an element which will make every group
        that contains it "fail".

    Returns
    -------
    selected_groups, error : tuple
        A tuple consisting of:
        - 'selected_groups', a dictionary of carefully selected elements from
        `possible_groups`, where keys are 'group' and values are [elements in group].
        The selected groups are such that knowing which of them failed or succeeded
        will be enough information to identify all bad elements in the original
        `possible_groups` set.
        - 'error', a string explaining what went wrong if there is an error in
        selection, otherwise it is None.
        If there is an error, then selected groups is an empty list, `[]`.
    """

    covering_elements = _minimal_elements_group_coverage(possible_groups)
    lcov = len(covering_elements)
    if lcov <= max_saboteurs:
        return (
            None,
            (
                "Not possible to detect up to %d saboteurs, as the following %d "
                "elements collectively cover all constructs, and therefore all "
                "constructs would fail at once if these %d elements alone were"
                "saboteurs:\n\n%s\n\n. Remove these elements from the problem or "
                " decrease the max number of saboteurs if possible."
            )
            % (max_saboteurs, lcov, lcov, ", ".join(covering_elements)),
        )
    all_elements = set(
        element for elements in possible_groups.values() for element in elements
    )

    product = itertools.product(*((1 + max_saboteurs) * (all_elements,)))
    all_tuples = set(tuple_ for tuple_ in product if len(set(tuple_)) == len(tuple_))

    def all_x_without_ys(group):
        return set(
            tuple_
            for tuple_ in all_tuples
            if (tuple_[0] in group) and not any((e in group) for e in tuple_[1:])
        )

    x_without_ys_sets = [
        (name, all_x_without_ys(group)) for name, group in possible_groups.items()
    ]
    selected = minimal_cover(all_tuples, x_without_ys_sets)
    if selected is None:
        return [], "No solution found."
    keys = list(possible_groups.keys())
    selected = sorted(selected, key=lambda group: keys.index(group))
    return OrderedDict((g, possible_groups[g]) for g in selected), None


def _groups_fail_table(groups):
    """Return a dict associating each element to the groups it can fail.

    This function is used both in ``plot_elements_in_group`` and
    ``find_logical_saboteurs``.

    Parameters
    ----------
    groups : OrderedDict
        Ordered dict of the form {group_name: [elements in groups]}.

    Returns
    -------
    fail_table : dict
        A dict {element_name: list[True False False ...]} where list[i] is True
        if the element is part of the i-th group in the provided ordered
        dictionary ``groups``.
    """

    groups = list(groups.values())
    all_elements = []
    for group in groups:
        for element in group:
            if element not in all_elements:
                all_elements.append(element)
    return OrderedDict(
        [
            (element, [(element in group) for group in groups])
            for element in all_elements
        ]
    )


def find_logical_saboteurs(groups, failed_groups):
    """Identify bad and suspicious elements from groups failure data.

    Parameters
    ----------
    groups : dict
        A dict {group_name: [elements in that group]}.
    failed_groups : list
        A list [group_name_1, group_name_2, ...] of the names of all groups that
        experimentally failed.

    Returns
    -------
    dict
        Returns a dictionary {'saboteurs': [...], 'suspicious': []} where
        ``suspicious`` is the list of all elements which do not appear in
        successful groups, and ``saboteurs`` is the list of suspicious elements
        which are also the only suspicious element in at least one group.
    """

    groups = OrderedDict(groups.items())
    failed_groups = set(failed_groups)
    groups_list = np.array(list(groups.keys()))
    fail_table = {
        element: set(groups_list[groups_])
        for element, groups_ in _groups_fail_table(groups).items()
    }
    suspicious = set(
        element
        for element, element_groups in fail_table.items()
        if element_groups < failed_groups
    )
    confirmed = set(
        suspect_element
        for suspect_element in suspicious
        if len(
            fail_table[suspect_element].difference(
                set().union(
                    *(
                        fail_table[other_suspect]
                        for other_suspect in suspicious.difference({suspect_element})
                    )
                )
            )
        )
    )
    return dict(
        saboteurs=list(confirmed), suspicious=list(suspicious.difference(confirmed))
    )



from collections import OrderedDict
import numpy as np
import matplotlib.pyplot as plt
import flametree
from .logical_methods import _groups_fail_table


def plot_batch(groups, ax=None):
    """Plot a diagram of all groups and the elements they contain.

    The ``groups`` parameter is a dict {group_name: [elements in the group]}.
    The ax is a Matplotlib Ax object on which to plot. If none is provided a
    new ax will be created and returned at the end.
    """
    groups = OrderedDict(groups.items())

    fail_table = _groups_fail_table(groups)
    array = np.array(list(fail_table.values())).T[::-1]
    lines, cols = len(array), len(array[0])

    if ax is None:
        _, ax = plt.subplots(1)
    ax.imshow(array, cmap="Purples")
    ax.set_aspect("equal")
    for y in range(lines):
        ax.axhline(y + 0.5, c="white")
    ax.set_yticks(range(lines))
    for x in range(cols):
        ax.axvline(x + 0.5, c="white")
    ax.set_xticks(range(cols))
    ax.set_xticklabels(list(fail_table.keys()), rotation=90)
    ax.set_yticklabels(list(groups.keys())[::-1])
    ax.set_xlim(-0.5, cols - 0.5)
    ax.set_ylim(-0.5, lines - 0.5)
    ax.set_aspect("equal")
    return ax


def generate_batch_report(
    groups, target="@memory", group_naming="group", plot_format="pdf"
):
    """
    Generate a report with CSV and plot describing a groups batch.

    Parameters
    ----------
    groups : OrderedDict
        A (ordered) dict {group_name: [elements in the group]}.
    target : str
        Either path to a folder, or a zip file, or "@memory" to return raw
        data of a zip file containing the report.
    group_naming : str
        Word that will replace "group" in the report, e.g. "assembly", "team", etc.
    plot_format : str
        Format of the plot (pdf, png, jpeg, etc).
    """

    root = flametree.file_tree(target)
    csv = ("%s,elements\n" % group_naming) + "\n".join(
        [",".join([group] + list(elements)) for group, elements in groups.items()]
    )
    root._file("%ss.csv" % group_naming).write(csv)
    ax = plot_batch(groups)
    ax.set_title("Elements per %s" % group_naming)
    ax.figure.savefig(
        root._file("%ss.%s" % (group_naming, plot_format)).open("wb"),
        bbox_inches="tight",
        format=plot_format,
    )
    return root._close()



from saboteurs import (
    generate_combinatorial_groups,
    design_test_batch,
    generate_batch_report,
)

elements_per_position = {
    "Position_1": ["A", "B", "C"],
    "Position_2": ["D", "E", "F", "G"],
    "Position_3": ["H", "I", "J", "K"],
    "Position_4": ["L", "M", "N"],
}
possible_groups = generate_combinatorial_groups(elements_per_position, prefix="G")
selected_groups, error = design_test_batch(possible_groups, max_saboteurs=2)
generate_batch_report(
    selected_groups, plot_format="png", target="design_test_batch_report"
)



import os
from saboteurs import csv_to_groups_data, find_statistical_saboteurs, statistics_report


def test_basics(tmpdir):
    csv_path = os.path.join("tests", "data", "statistical.csv")
    pdf_path = os.path.join(str(tmpdir), "test.pdf")
    groups_data = csv_to_groups_data(csv_path)
    analysis_results = find_statistical_saboteurs(groups_data)
    statistics_report(analysis_results, pdf_path)
    data = statistics_report(analysis_results, "@memory")
    assert len(data) > 50000



import os
from saboteurs import (
    find_logical_saboteurs,
    generate_combinatorial_groups,
    design_test_batch,
    csv_to_groups_data,
    generate_batch_report,
)


def test_find_logical_saboteurs():
    groups = {
        1: ["A", "C", "D"],
        2: ["B", "C", "E"],
        3: ["A", "B", "D"],
        4: ["D", "F", "G"],
    }
    result = find_logical_saboteurs(groups, failed_groups=[2, 4])
    assert result["saboteurs"] == ["E"]
    assert sorted(result["suspicious"]) == ["F", "G"]


def test_find_logical_saboteurs_from_csv():
    csv_path = os.path.join("tests", "data", "logical.csv")
    groups, failed_groups = csv_to_groups_data(csv_path)
    result = find_logical_saboteurs(groups, failed_groups=failed_groups)
    assert result["saboteurs"] == ["E"]
    assert sorted(result["suspicious"]) == ["F", "G"]


def test_design_test_batch():
    elements_per_position = {
        "Position_1": ["A", "B", "C"],
        "Position_2": ["D", "E", "F", "G"],
        "Position_3": ["H", "I", "J", "K"],
        "Position_4": ["L", "M", "N"],
    }
    possible_groups = generate_combinatorial_groups(elements_per_position)
    selected_groups, error = design_test_batch(possible_groups, max_saboteurs=2)
    assert error is None
    assert len(selected_groups) == 15
    raw_data = generate_batch_report(selected_groups)
    assert len(raw_data) > 2000



from saboteurs import csv_to_groups_data, find_statistical_saboteurs, statistics_report

groups_data = csv_to_groups_data("assemblies_data.csv")
analysis_results = find_statistical_saboteurs(groups_data)
statistics_report(
    analysis_results,
    "genetic_parts_example.pdf",
    replacements=[("groups", "assemblies"), ("group", "assembly"), ("member", "part")],
)



from collections import OrderedDict


def csv_to_groups_data(csv_path=None, csv_string=None):
    """Read a CSV to get the data to feed to ``find_statistical_saboteurs()``
    or ``find_logical_saboteurs()``.

    See examples of such a file in the code repository:
    https://github.com/Edinburgh-Genome-Foundry/saboteurs/

    Returns
    -------
    groups, failed_groups : tuple
      For datasheets for logical saboteur finding.
    group_data : dict
      For datasheets for statistical saboteur finding. The data is of the form:

      >>> {"Exp. 1": {
      >>>      exp_id: "Exp. 1",
      >>>     attempts: 7,
      >>>     failures: 10,
      >>>     members: ["Alice", "Bob"]}
      >>>  }
      >>>  "Exp. 2": { etc...
    """
    if csv_string is None:
        with open(csv_path, "r") as f:
            csv_string = f.read()
    lines = [
        [e.strip() for e in l.split(",") if len(e.strip())]
        for l in csv_string.split("\n")
        if len(l)
    ]
    groups = OrderedDict([])
    if "result" in lines[0]:
        failed_groups = []
        for line in lines[1:]:
            name, result, members = line[0], line[1], line[2:]
            groups[name] = members
            if result != "success":
                failed_groups.append(name)
        return groups, failed_groups
    else:
        for line in lines[1:]:
            (name, attempts, failures), members = line[:3], line[3:]
            groups[name] = dict(
                id=name, attempts=int(attempts), failures=int(failures), members=members
            )
        return groups
